//*****************************************************************
// Questo libreria si occupa di implementare la funzione di una 
// unit di controllo all'interno di un una macchina tipo quella di 
// Von Neumann e rappresenta dunque il cuore vero e proprio di 
// tutto il sistema.
//
// Nome del progetto     : control_unit.h control_unit.cpp
// Creatore              : Giovanni Ceglia
// Data ultima revisione : 16/05/2001
//
// Librerie e files richiesti:    io_unit.h io_unit.cpp
//                                vmachine.h vmachine.cpp
//                                tplus.h tplus.cpp
//                                alu_unit.h alu_unit.cpp
//*****************************************************************
 
#include "tplus.h"         // libreria che contiene implementazioni
                           // di tipi nuovi, come quello stringa
                           // oppure il tipo boolean e relative 
                           // funzioni utili alla gestione.
                           
#include "vmachine.h"      // libreria che implementa la macchina virtuale

#include "alu_unit.h"      // simulazione della ALU
#include "io_unit.h"       // unita di input output



// funzione che va in ciclo e che simula il funzionamento vero e prioprio di
// una ipotetica unit di controllo.
void unita_di_controllo(virtual_machine&);

// fase di caricamento e di preparazione all'esecuzione della prossima 
// istruzione
void fase_fetch(virtual_machine&);

// esecuzione dell'istruzione prelevata dalla memoria
void fase_execute(virtual_machine&);

// unit specifica che si occupa di implementare ed eseguire
// le funzioni proprie dell ' unit di controllo.
void uc_exec(virtual_machine&);

void fase_fetch(virtual_machine& vm)
 {
  int opcodes;
  // Calcolo della dimensione in bytes dell'istruzione.
  opcodes=(vm.virtual_memory[vm.virtual_PI] / 64)+1;
  // Caricamento dell ' istruzione nel registro apposito. 
  vm.reg_istruzione=vm.virtual_memory[vm.virtual_PI];
  // Caricamento degli eventuali operandi.
  if (opcodes==2) vm.reg_opcode_A=vm.virtual_memory[vm.virtual_PI+1];
   else if (opcodes==3) 
     {
      vm.reg_opcode_A=vm.virtual_memory[vm.virtual_PI+1];
      vm.reg_opcode_B=vm.virtual_memory[vm.virtual_PI+2];
     }    
  // incremento del Puntatore Prossima Istruzione PI. 
  vm.virtual_PI=vm.virtual_PI+opcodes;
 }

void uc_exec(virtual_machine& vm)
 {
   char tmp;
   int i;
   switch (vm.reg_istruzione) {
    // Istruzione di RESET, azzera completamente lo stato della macchina.
     case (RST) : {
                  vm.virtual_flags=0;
                  vm.virtual_PI=0;
                  for (i=0;i<256;i++) vm.virtual_memory[i]=0;
                  vm.reg_istruzione=0;
                  vm.reg_opcode_A=0;
                  vm.reg_opcode_B=0;
                  vm.reg_offset=0;
                } break;
    // Istruzioni per settare ad uno i flags di stato.
     case STD :   vm.virtual_flags=vm.virtual_flags | 0x01; break;
     case STZ :   vm.virtual_flags=vm.virtual_flags | 0x02; break;
     case STO :   vm.virtual_flags=vm.virtual_flags | 0x04; break;
     case STP :   vm.virtual_flags=vm.virtual_flags | 0x08; break;
     case STN :   vm.virtual_flags=vm.virtual_flags | 0x10; break;
    // Istruzione di nessuna operazione.  
     case NOP :   break;
    // Istruzioni per settare a zero i flags di stato.  
     case CLD :   vm.virtual_flags=vm.virtual_flags & 0x1E; break;
     case CLZ :   vm.virtual_flags=vm.virtual_flags & 0x1D; break;
     case CLO :   vm.virtual_flags=vm.virtual_flags & 0x1B; break;
     case CLP :   vm.virtual_flags=vm.virtual_flags & 0x17; break;
     case CLN :   vm.virtual_flags=vm.virtual_flags & 0x0F; break;
    // Istruzioni di salto.  
     case JMP : {
                  vm.virtual_PI=vm.reg_opcode_A;
                } break;
     case JPD : {
                  if ((vm.virtual_flags & 0x01)>0) vm.virtual_PI=vm.reg_opcode_A;
                } break;
     case JPZ : {
                  if ((vm.virtual_flags & 0x02)>0) vm.virtual_PI=vm.reg_opcode_A;
                } break;
     case JPO : {
                  if ((vm.virtual_flags & 0x04)>0) vm.virtual_PI=vm.reg_opcode_A;
                } break;
     case JPP : {
                  if ((vm.virtual_flags & 0x08)>0) vm.virtual_PI=vm.reg_opcode_A;
                } break;
     case JPN : {
                  if ((vm.virtual_flags & 0x10)>0) vm.virtual_PI=vm.reg_opcode_A;
                } break;
    // Istruzione di copia di valori.  
     case CPY : {
                  tmp=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
                  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=vm.virtual_memory[vm.reg_opcode_B];
                  vm.virtual_memory[vm.reg_opcode_B]=tmp;
                  if (vm.reg_opcode_B==0) vm.virtual_flags=vm.virtual_flags | 0x01;
                   else if (vm.reg_opcode_B<0) vm.virtual_flags=vm.virtual_flags | 0x10;
                    else vm.virtual_flags=vm.virtual_flags | 0x08;
                } break;
     case VAL : {
                  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=vm.reg_opcode_B;
                  if (vm.reg_opcode_B==0) vm.virtual_flags=vm.virtual_flags | 0x01;
                   else if (vm.reg_opcode_B<0) vm.virtual_flags=vm.virtual_flags | 0x10;
                    else vm.virtual_flags=vm.virtual_flags | 0x08;
                } break;
     case VALOFFST: vm.reg_offset=vm.reg_opcode_A; break;
     default : break;
    }  
 }
 
 void fase_execute(virtual_machine& vm)
 {
   switch (vm.reg_istruzione) {
      case RST : uc_exec(vm); break;
      case STD : uc_exec(vm); break;
      case STZ : uc_exec(vm); break;
      case STO : uc_exec(vm); break;
      case STP : uc_exec(vm); break;
      case STN : uc_exec(vm); break;
      case NOP : uc_exec(vm); break;
      case CLD : uc_exec(vm); break;
      case CLZ : uc_exec(vm); break;
      case CLO : uc_exec(vm); break;
      case CLP : uc_exec(vm); break;
      case CLN : uc_exec(vm); break;
      case JMP : uc_exec(vm); break;
      case JPD : uc_exec(vm); break;
      case JPZ : uc_exec(vm); break;
      case JPO : uc_exec(vm); break;
      case JPP : uc_exec(vm); break;
      case JPN : uc_exec(vm); break;
      case CPY : uc_exec(vm); break;
      case VAL : uc_exec(vm); break;
      case VALOFFST: uc_exec(vm); break;
      case INCOFFST: alu_exec(vm); break;
      case DECOFFST: alu_exec(vm); break;
      // operazioni di Input Output      
      case INP : io_exec(vm); break;
      case WRT : io_exec(vm); break;      
      // operazioni destinate all'unit ALU (Unit Logico Aritmetica)
      case INC : alu_exec(vm); break;
      case DEC : alu_exec(vm); break;
      case CMP : alu_exec(vm); break;
      case ADD : alu_exec(vm); break;
      case SUB : alu_exec(vm); break;
      case DIV : alu_exec(vm); break;
      case MUL : alu_exec(vm); break;
      default  : break;
    }
 }


void unita_di_controllo(virtual_machine& vm)
 {
  while (vm.virtual_memory[vm.virtual_PI]!=RST){
     fase_fetch(vm);
     if ((vm.virtual_flags & 0x01)>0) visualizza_stato_macchina(vm);
     fase_execute(vm);
   }
 }



